--[[

Skillet: A tradeskill window replacement.
Copyright (c) 2007 Robert Clark <nogudnik@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

]]


local AceEvent = AceLibrary("AceEvent-2.0")
local PT
if AceLibrary:HasInstance("LibPeriodicTable-3.1") then
	PT = AceLibrary("LibPeriodicTable-3.1")
end

-- a table of tradeskills by id (lowest qualifying skill only)
local TradeSkillList = {
	2259,           -- alchemy
	2018,           -- blacksmithing
	7411,           -- enchanting 
	4036,           -- engineering
	25229,          -- jewelcrafting
	2108,           -- leatherworking
--	2575,			-- mining (or smelting?)
	2656,           -- smelting (from mining)
	3908,           -- tailoring
	2550,           -- cooking
	3273,           -- first aid
	2842,           -- poisons
	
--	5149, 			-- beast training (not supported, but i need to know the number)
}


local TradeSkillIDsByName = {}		-- filled in with ids and names for reverse matching (since the same name has multiple id's based on level)

local DifficultyText = {
	x = "unknown",
	o = "optimal",
	m = "medium",
	e = "easy",
	t = "trivial",
}
local DifficultyChar = {
	unknown = "x",
	optimal = "o",
	medium = "m",
	easy = "e",
	trivial = "t",
}


local skill_style_type = {
	["unknown"]			= { r = 1.00, g = 0.00, b = 0.00, level = 5, alttext="???"},
	["optimal"]	        = { r = 1.00, g = 0.50, b = 0.25, level = 4, alttext="+++"},
	["medium"]          = { r = 1.00, g = 1.00, b = 0.00, level = 3, alttext="++"},
	["easy"]            = { r = 0.25, g = 0.75, b = 0.25, level = 2, alttext="+"},
	["trivial"]	        = { r = 0.50, g = 0.50, b = 0.50, level = 1, alttext=""},
	["header"]          = { r = 1.00, g = 0.82, b = 0,    level = 0, alttext=""},
}



-- adds an recipe source for an itemID (recipeID produces itemID)
function Skillet:ItemDataAddRecipeSource(itemID,recipeID)
	if not itemID or not recipeID then return end
	
	if not self.data.itemRecipeSource then
		self.data.itemRecipeSource = {}
	end
	
	if not self.data.itemRecipeSource[itemID] then		
		self.data.itemRecipeSource[itemID] = {}
	end
	
	self.data.itemRecipeSource[itemID][recipeID] = true
end


-- adds a recipe usage for an itemID (recipeID uses itemID as a reagent)
function Skillet:ItemDataAddUsedInRecipe(itemID,recipeID)
	if not itemID or not recipeID then return end
	
	if not self.data.itemRecipeUsedIn then
		self.data.itemRecipeUsedIn = {}
	end
	
	if not self.data.itemRecipeUsedIn[itemID] then
		self.data.itemRecipeUsedIn[itemID] = {}
	end
	
	self.data.itemRecipeUsedIn[itemID][recipeID] = true
end


-- goes thru the stored recipe list and collects reagent and item information as well as skill lookups
function Skillet:CollectRecipeInformation()
	
--[[
	for recipeID in pairs(self.db.account.recipeDB) do
		local recipe = self:GetRecipe(recipeID)
		
		if recipe.itemID ~= 0 then
			self:ItemDataAddRecipeSource(recipe.itemID, recipeID)
		end
		
		local reagentData = recipe.reagentData

		if reagentData then
			for r=1,#reagentData do
				self:ItemDataAddUsedInRecipe(reagentData[r].id, recipeID)
			end
		end
	end
]]
	for recipeID, recipeString in pairs(self.db.account.recipeDB) do
		local recipe = self:GetRecipe(recipeID)
		
		local tradeID, itemString, reagentString, toolString = string.split(" ",recipeString)
		local itemID, numMade = 0, 1
		local slot = nil
		
		if itemString ~= "0" then
			local a, b = string.split(":",itemString)
			
			if a ~= "0" then 
				itemID, numMade = a,b
			else
				itemID = 0
				numMade = 1
				slot = tonumber(b)
			end
			
			if not numMade then
				numMade = 1
			end
		end
		
		itemID = tonumber(itemID)
		
		if itemID ~= 0 then
			self:ItemDataAddRecipeSource(itemID, recipeID)
		end
		

		local reagentList = { string.split(":",reagentString) }
		local numReagents = #reagentList / 2
		
		for i=1,numReagents do
			local reagentID = tonumber(reagentList[1 + (i-1)*2])
			
			self:ItemDataAddUsedInRecipe(reagentID, recipeID)
		end
	end
	
	
	for player,tradeList in pairs(self.db.server.skillDB) do
		self.data.skillIndexLookup[player] = {}
		
		for trade,skillList in pairs(tradeList) do
			for i=1,#skillList do
--				local skillData = self:GetSkill(player, trade, i)
				local skillString = self.db.server.skillDB[player][trade][i]

				local data = { string.split(" ",skillString) }
				
				if data[1] ~= "header" then
					local recipeID = tonumber(string.sub(data[1],2))
				
					self.data.skillIndexLookup[player][recipeID] = i
				end
			end
		end
	end
end



-- Checks to see if the current trade is one that we support.

function Skillet:IsSupportedTradeskill(tradeID)
	if not tradeID or tradeID == 5419 then
		return false				-- beast training
	end
	
	if IsShiftKeyDown() then
		return false
	end
	
	-- EnchantingSell does not play well with the Skillet window, so
	-- if it is enabled, and it was the craft frame hidden, do not
	-- show Skillet for enchanting.
	--
	-- EnchantingSell does some odd things to the enchanting toggle,
	-- so expect some odd bug reports about this.
	if self:IsCraft() and ESeller and ESeller:IsActive() and ESeller.db.char.DisableDefaultCraftFrame then
		 return false
	end
	
	return true
end


-- queries periodic table for vendor info for a particual itemID
function Skillet:VendorSellsReagent(itemID)
	if PT then
		-- 30817 == simple flour
		-- 4539 == Goldenbark Apple
		-- 17035 == Stranglethorn seed
		-- 17034 == Maple Seed
		
		local missingItems = {
			[30817] = true,				-- simple flour
			[4539] = true,				-- Goldenbark Apple
			[17035] = true,				-- Stranglethorn seed
			[17034] = true, 			-- Maple seed
		}
		
		if missingItems[itemID] then
			return true
		end
		
		if PT:ItemInSet(itemID,"Tradeskill.Mat.BySource.Vendor") then
			return true
		end
	end
end

-- resets the blizzard tradeskill search filters just to make sure no other addon has monkeyed with them
function Skillet:ResetTradeSkillFilter()
	SetTradeSkillItemNameFilter("")	 			
	SetTradeSkillItemLevelFilter(0,0)
end

-- resets the crafting search filter
function Skillet:ResetCraftFilter()
	SetCraftItemNameFilter("")
--	SetCraftItemFilter(0)
end



local lasGetRecipeID = nil
-- reconstruct a recipe from a recipeString and cache it into our system for this session
function Skillet:GetRecipe(id)
	if not id or id == 0 then return self.unknownRecipe end
--DEFAULT_CHAT_FRAME:AddMessage("recipe: "..id)
	
--	if id == lastGetRecipeID then
--		return self.data.recipeList[1]
--	end
	
	lastGetRecipeID = id
	
	if (not self.data.recipeList[id]) and self.db.account.recipeDB[id] then	
		local recipeString = self.db.account.recipeDB[id]
		
--		id = 1
		
		local tradeID, itemString, reagentString, toolString = string.split(" ",recipeString)
		local itemID, numMade = 0, 1
		local slot = nil
		
		if itemString ~= "0" then
			local a, b = string.split(":",itemString)
			
			if a ~= "0" then 
				itemID, numMade = a,b
			else
				itemID = 0
				numMade = 1
				slot = tonumber(b)
			end
			
			if not numMade then
				numMade = 1
			end
		end
		
		self.data.recipeList[id] = {}
		
		self.data.recipeList[id].tradeID = tonumber(tradeID)
		self.data.recipeList[id].itemID = tonumber(itemID)
		self.data.recipeList[id].numMade = tonumber(numMade)
		
		self.data.recipeList[id].slot = slot
		
		self.data.recipeList[id].reagentData = {}

		local reagentList = { string.split(":",reagentString) }
		local numReagents = #reagentList / 2
		
		for i=1,numReagents do
			self.data.recipeList[id].reagentData[i] = {}
			
			self.data.recipeList[id].reagentData[i].id = tonumber(reagentList[1 + (i-1)*2])
			self.data.recipeList[id].reagentData[i].numNeeded = tonumber(reagentList[2 + (i-1)*2])
		end
		
		if toolString ~= "-" then
			self.data.recipeList[id].tools = {}
		
			local toolList = { string.split(":",toolString) }

			for i=1,#toolList do
				self.data.recipeList[id].tools[i] = string.gsub(toolList[i],"_"," ")
			end
		end
	end
	
	return self.data.recipeList[id] or self.unknownRecipe
end


-- reconstruct a skill from a skillString and cache it into our system for this session
function Skillet:GetSkill(player,trade,index)
	if player and trade and index then
--DebugSpam(key)
		if not self.data.skillList[player] then
			self.data.skillList[player] = {}
		end
		
		if not self.data.skillList[player][trade] then
			self.data.skillList[player][trade] = {}
		end
		
		if not self.data.skillList[player][trade][index] then	
			local skillString = self.db.server.skillDB[player][trade][index]
--DebugSpam(skillString)			
			if skillString then
				local skill = {}
				
				local data = { string.split(" ",skillString) }
				
				if data[1] == "header" then
					skill.id = 0
				else
					local difficulty = string.sub(data[1],1,1)
					local recipeID = tonumber(string.sub(data[1],2))
					
					skill.id = recipeID
					skill.difficulty = DifficultyText[difficulty]
					skill.color = skill_style_type[DifficultyText[difficulty]]
					skill.tools = nil
					
					for i=2,#data do
						local subData = { string.split("=",data[i]) }
						
						if subData[1] == "cd" then
							skill.cooldown = tonumber(subData[2])
--DEFAULT_CHAT_FRAME:AddMessage(skill.cooldown)
						elseif subData[1] == "t" then
							local recipe = self:GetRecipe(recipeID)
						
							skill.tools = {}
							
							for j=1,string.len(subData[2]) do
								local missingTool = tonumber(string.sub(subData[2],j,j))
								skill.tools[missingTool] = true
							end
						end
					end
				end
				
				self.data.skillList[player][trade][index] = skill
			end
		end
		
		return self.data.skillList[player][trade][index]
	end
end


-- collects generic tradeskill data (id to name and name to id)
function Skillet:CollectTradeSkillData()
	for i=1,#TradeSkillList,1 do
		local id = TradeSkillList[i]
		local name, _, icon = GetSpellInfo(id)
		
		TradeSkillIDsByName[name] = id
	end

	self.tradeSkillIDsByName = TradeSkillIDsByName
	self.tradeSkillList = TradeSkillList
end

-- this routine collects the basic data (which tradeskills a player has)
-- clean = true means wipe the old data
function Skillet:ScanPlayerTradeSkills(player, clean)
DebugSpam("ScanPlayerTradeSkills "..(player or "nil"))	
	if player == (UnitName("player")) then			            -- only for active player

		if clean or not Skillet.db.server.skillRanks[player] then
			Skillet.db.server.skillRanks[player] = {}
		end
		
		local skillRanksData = Skillet.db.server.skillRanks[player]
		
		for i=1,#TradeSkillList,1 do

			local id = TradeSkillList[i]
			local name = GetSpellInfo(id)			            -- always returns data
			local _, rankName, icon = GetSpellInfo(name)		    -- only returns data if you have this spell in your spellbook
	
			
DebugSpam("collecting tradeskill data for "..name.." "..(rank or "nil"))

			if rankName then
				if not skillRanksData[id] then
					skillRanksData[id] = ""
				end
			else
				skillRanksData[id] = nil	
			end
		end
	end
end


function tsithing()
-- this should be broken out into an addon, but for now i'm leaving it here to help me figure out how to configure such an addon
	if TradeskillInfo then -- tsi is installed
		local order = 1
		player = "tsiData"
		
		Skillet.db.server.skillRanks[player] = {}
		Skillet.db.server.skillData[player] = {}
		
		for i,skillName in pairs(TradeskillInfo.vars.tradeskills) do
			
			local newSkill = {}
			newSkill.rank = 375
			newSkill.maxRank = 375
			newSkill.texture = GetSpellTexture(self:GetSpellByProfession(skillName))
			
			-- hacky way to get things to be in a nice order for menus
			-- while retaining random access to database
			newSkill.order = order

			self.db.server.skillRanks[player][skillName] = newSkill


			order = order + 1
			
			self.db.server.skillData[player][skillName] = {}
		end
	end

end


function Skillet:EnableDataGathering(addon)
	assert(tostring(addon),"Usage: EnableDataGathering('addon')")
--	self.datagatheraddons[addon] = true
	self:RegisterEvent("CHAT_MSG_SKILL")
	self:RegisterEvent("CHAT_MSG_SYSTEM")
	self:RegisterEvent("TRADE_SKILL_UPDATE")
	
	self.dataScanned = false
	
	self:CollectTradeSkillData()
		
	self:RegisterPlayerDataGathering((UnitName("player")),self)		-- tell self who we want to collect data for and how to do it
end


-- not sure this ever really needs to be called...
function Skillet:DisableDataGathering(addon)
--[[
	if not addon then
		self.data = nil
		self.datagatheraddons = {}
		return
	end
	assert(tostring(addon),"Usage: DisableDataGathering(['addon'])")
	self.datagatheraddons[addon] = false
	if next(self.datagatheraddons) then
		return
	end
]]

--	self:UnregisterEvent("TRADE_SKILL_SHOW")
--	self:UnregisterEvent("CRAFT_SHOW")
	self:UnregisterEvent("CHAT_MSG_SKILL")
	self:UnregisterEvent("CHAT_MSG_SYSTEM")
	
	self.data = nil
end


function Skillet:EnableQueue(addon)
	assert(tostring(addon),"Usage: EnableDataGathering('addon')")
--	self.queueaddons[addon] = true
	self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED",   "StopCastCheckUnit")
	self:RegisterEvent("UNIT_SPELLCAST_FAILED",      "StopCastCheckUnit")
	self:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED", "StopCastCheckUnit")
	self:RegisterEvent("UNIT_SPELLCAST_STOPPED",     "StopCastCheckUnit")
--	if not self.queue then
--		self.queue = {}
--	end
--	self.queueenabled = true
end


function Skillet:DisableQueue(addon)
--[[	if not addon then
		self.queue = nil
		self.queueaddons = {}
		self.queueenabled = false
		return
	end
	assert(tostring(addon),"Usage: DisableDataGathering(['addon'])")
	self.queueaddons[addon] = false
	if next(self.queueaddons) then
		return
	end
]]--
	self:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	self:UnregisterEvent("UNIT_SPELLCAST_FAILED")
	self:UnregisterEvent("UNIT_SPELLCAST_INTERRUPTED")
	self:UnregisterEvent("UNIT_SPELLCAST_STOPPED")
--	self.queueenabled = false
--	self.queue = nil
end


-- takes a profession and a skill index and returns the recipe
function Skillet:GetRecipeDataByTradeIndex(tradeID, index)
	if not tradeID or not index then
		return self.unknownRecipe
	end
	
	local skill = self:GetSkill(self.currentPlayer, tradeID, index)
	
	if skill then 
		local recipeID = skill.id
		
		if recipeID then
	--		local recipeData = self.db.account.recipeData[recipeID] or selfUnknownRecipe
			local recipeData = self:GetRecipe(recipeID)
			
			return recipeData, recipeID
		end
	end
end


function Skillet:StopCastCheckUnit(unit, spell, rank)
DebugSpam(event.." "..(unit or "nil"))
	if unit == "player" then
		self:StopCast(spell)
--		AceEvent:ScheduleEvent("Skillet_StopCast", self.StopCast, 0.1,self,event,spell)
	end
end


-- Internal
function Skillet:Skillet_AutoRescan()
DebugSpam("AUTO RESCAN")
	if InCombatLockdown() then
		return
	end

	if AceEvent:IsEventScheduled("Skillet_AutoRescan") then
		AceEvent:CancelScheduledEvent("Skillet_AutoRescan")
	end

	self:RescanTrade()
	
	self:UpdateTradeSkillWindow()
DebugSpam("AUTO RESCAN COMPLETE")
end


function Skillet:TRADE_SKILL_UPDATE()
	if AceEvent:IsEventScheduled("Skillet_AutoRescan") then
		AceEvent:CancelScheduledEvent("Skillet_AutoRescan")
	end
	
	AceEvent:ScheduleEvent("Skillet_AutoRescan", self.Skillet_AutoRescan, 0.5,self)
end


function Skillet:CHAT_MSG_SKILL()
DebugSpam("CHAT_MSG_SKILL "..(event or "nil"))
--	self:Skillet_AutoRescan()									-- the problem here is that the message comes before the actuality, it seems
	if AceEvent:IsEventScheduled("Skillet_AutoRescan") then
		AceEvent:CancelScheduledEvent("Skillet_AutoRescan")
	end
	
	AceEvent:ScheduleEvent("Skillet_AutoRescan", self.Skillet_AutoRescan, 0.5,self)
end

function Skillet:CHAT_MSG_SYSTEM()
	local cutString = string.sub(1,(string.find(ERR_LEARN_RECIPE_S,"%s")))
--DebugSpam("CHAT_MSG_SYSTEM "..(arg1 or "nil").." vs "..cutString)
	if arg1 and string.find(arg1, cutString) then
--		self:Skillet_AutoRescan()								-- the problem here is that the message comes before the actuality, it seems
		if AceEvent:IsEventScheduled("Skillet_AutoRescan") then
			AceEvent:CancelScheduledEvent("Skillet_AutoRescan")
		end

		AceEvent:ScheduleEvent("Skillet_AutoRescan", self.Skillet_AutoRescan, 0.5,self)
	end
end



function Skillet:CalculateCraftableCounts(playerOverride)
DebugSpam("CalculateCraftableCounts")
	local player = playerOverride or self.currentPlayer
	local skillDB = self.db.server.skillDB[player][self.currentTrade]
DebugSpam((player or "nil").." "..(self.currentTrade or "nil"))

DebugSpam("recalculating crafting counts")
	self.visited = {}
	
	if skillDB then
	DebugSpam("has skillDB")
		for i=1,#skillDB,1 do
			local skill = self:GetSkill(player, self.currentTrade, i)
			
			if skill then			-- skip headers
				skill.numCraftable, skill.numCraftableVendor, skill.numCraftableBank, skill.numCraftableAlts = self:InventorySkillIterations(self.currentTrade, i, player)
			end
		end
	end
DebugSpam("CalculateCraftableCounts Complete")
end


local scanInProgress = false;


function Skillet:ScanTrade()
DebugSpam("ScanTrade")
	if self.scanInProgress == true then
DebugSpam("SCAN BUSY!")
		return
	end
	
	self.scanInProgress = true
	
	local tradeID
	
	local API = {}
	
	local profession, rank, maxRank = GetTradeSkillLine()
DebugSpam("GetTradeSkill: "..(profession or "nil"))

	if not profession or profession == "UNKNOWN" then 
		profession, rank, maxRank = GetCraftDisplaySkillLine()
		
		DebugSpam("GetTradeSkill: "..(profession or "nil"))
		
		if not profession then
			self.scanInProgress = false
			return false
		end
		
		if CraftIsPetTraining() then			-- new function in 2.2 to check if the current craft is pet training
			self.scanInProgress = false
			return false
		end
	
		tradeID = 7411					-- enchanting... hardwired for now since there are no others of interest to us
		
		API.GetNumSkills = GetNumCrafts
		API.ExpandLine = ExpandCraftLine
		API.GetRecipeLink = GetCraftRecipeLink
		API.GetTools = GetCraftSpellFocus
		API.GetCooldown = GetCraftCooldown
		API.GetItemLink = GetCraftItemLink
		API.GetNumMade = GetCraftNumMade
		API.GetNumReagents = GetCraftNumReagents
		API.GetReagentInfo = GetCraftReagentInfo
		API.GetReagentLink = GetCraftReagentItemLink
	else
		API.GetNumSkills = GetNumTradeSkills
		API.ExpandLine = ExpandTradeSkillSubClass
		API.GetRecipeLink = GetTradeSkillRecipeLink
		API.GetTools = GetTradeSkillTools
		API.GetCooldown = GetTradeSkillCooldown
		API.GetItemLink = GetTradeSkillItemLink
		API.GetNumMade = GetTradeSkillNumMade
		API.GetNumReagents = GetTradeSkillNumReagents
		API.GetReagentInfo = GetTradeSkillReagentInfo
		API.GetReagentLink = GetTradeSkillReagentItemLink
		
		-- get the tradeID from the profession name (data collected earlier).
		tradeID = TradeSkillIDsByName[profession] or 2656				-- "mining" doesn't exist as a spell, so instead use smelting (id 2656)
	end
	
	
	if tradeID ~= self.currentTrade then
DebugSpam("TRADE MISMATCH for player "..(self.currentPlayer or "nil").."!  "..(tradeID or "nil").." vs "..(self.currentTrade or "nil"));
	end

	
	local player = self.currentPlayer
	
	if not self.recacheRecipe then
		self.recacheRecipe = {}
	end
	

	self.db.server.skillRanks[player][tradeID] = rank.." "..maxRank
	
	
	self:ResetTradeSkillFilter()						-- verify the search filter is blank (so we get all skills)
	self:ResetCraftFilter()
	
	
	local numSkills = API.GetNumSkills()
	

DebugSpam("Scanning Trade "..(profession or "nil")..":"..(tradeID or "nil").." "..numSkills.." recipes")

	if not self.data.skillIndexLookup[player] then
		self.data.skillIndexLookup[player] = {}
	end
	
	local skillDB = self.db.server.skillDB[player][tradeID]
	local skillData = self.data.skillList[player][tradeID]
	local recipeDB = self.db.account.recipeDB
	
	local lastHeader = nil
	local gotNil = false
	
	local currentGroup = nil
	
	local mainGroup = self:RecipeGroupNew(player,tradeID,"Blizzard")
	
	mainGroup.locked = true
	
	self:RecipeGroupClearEntries(mainGroup)
	
	local groupList = {}
	
	
	for i = 1, numSkills, 1 do
		repeat
DebugSpam("scanning index "..i)
			local skillName, skillType, isExpanded, subSpell, extra
			
			if tradeID == 7411 then
				skillName, subSpell, skillType, _, isExpanded, extra = GetCraftInfo(i)
			else
				skillName, skillType, _, isExpanded = GetTradeSkillInfo(i)
			end
			
			
DebugSpam("**** skill: "..(skillName or "nil"))

			gotNil = false
		
			
			if skillName then
				if skillType == "header" then
					if not isExpanded then
						API.ExpandLine(i)
					end

					local groupName
					
					if groupList[skillName] then
						groupList[skillName] = groupList[skillName]+1
						groupName = skillName.." "..groupList[skillName]
					else
						groupList[skillName] = 1
						groupName = skillName
					end
					
					skillDB[i] = "header "..skillName
					skillData[i] = nil
					
					currentGroup = self:RecipeGroupNew(player, tradeID, "Blizzard", groupName)
					
					self:RecipeGroupAddSubGroup(mainGroup, currentGroup, i)
				else
					local recipeLink = API.GetRecipeLink(i)
					local recipeID = self:GetItemIDFromLink(recipeLink)
					
					if not recipeID then
						gotNil = true
						break
					end
					
					if currentGroup then
						self:RecipeGroupAddRecipe(currentGroup, recipeID, i)
					else
						self:RecipeGroupAddRecipe(mainGroup, recipeID, i)
					end
					
					
					-- break recipes into lists by profession for ease of sorting
					skillData[i] = {}
				
	--					skillData[i].name = skillName
					skillData[i].id = recipeID					
					skillData[i].difficulty = skillType
					skillData[i].color = skill_style_type[skillType]
	--				skillData[i].category = lastHeader
					
					
					local skillDBString = DifficultyChar[skillType]..recipeID
					
					
					local tools = { API.GetTools(i) }

					skillData[i].tools = {}
					
					local slot = 1
					for t=2,#tools,2 do
						skillData[i].tools[slot] = (tools[t] or 0)						
						slot = slot + 1
					end
					
					local cd = API.GetCooldown(i)
					
					if cd then
						skillData[i].cooldown = cd + time()		-- this is when your cooldown will be up
				
						skillDBString = skillDBString.." cd=" .. cd + time()
					end

					local numTools = #tools+1
					
					if numTools > 1 then
						local toolString = ""
						local toolsAbsent = false
						local slot = 1
						
						for t=2,numTools,2 do
							if not tools[t] then
								toolsAbsent = true
								toolString = toolString..slot
							end
							
							slot = slot + 1
						end
						
						if toolsAbsent then										-- only point out missing tools
							skillDBString = skillDBString.." t="..toolString
						end
					end
					
					skillDB[i] = skillDBString
					
					self.data.skillIndexLookup[self.currentPlayer][recipeID] = i
					
					if recipeDB[recipeID] and not self.recacheRecipe[recipeID] then
						-- presumably the data is the same, so there's not much that needs to happen here.
						-- potentially, however, i could see an instance where a mod might feed tradeskill info and then "better" tradeskill info
						-- might be retrieved from the server which should over-ride the earlier tradeskill info
						-- (eg, tradeskillinfo sends skillet some data and then we learn that data was not quite up-to-date)

					else
						self.data.recipeList[recipeID] = {}
						
						local recipe = self.data.recipeList[recipeID]
						local recipeString
						local toolString = "-"
						
						recipe.tradeID = tradeID
						
						
						
						if #tools >= 1 then
							recipe.tools = { tools[1] }
							
							toolString = string.gsub(tools[1]," ", "_")
					
							for t=3,#tools,2 do
								table.insert(recipe.tools, tools[t])
								toolString = toolString..":"..string.gsub(tools[t]," ", "_")
							end
							
						end
						
						
						local itemLink = API.GetItemLink(i)
						
						if not itemLink then
							gotNil = true
							break
						end
						
						local itemString = "0"
						
						if GetItemInfo(itemLink) then
							local itemID = self:GetItemIDFromLink(itemLink)
							
							local minMade,maxMade = API.GetNumMade(i)
						
							recipe.itemID = itemID
							recipe.numMade = (minMade + maxMade)/2
							
							if recipe.numMade > 1 then
								itemString = itemID..":"..recipe.numMade
							else
								itemString = itemID
							end
							
							self:ItemDataAddRecipeSource(itemID,recipeID)					-- add a cross reference for the source of particular items
						else
							recipe.numMade = 1												
							recipe.itemID = 0												-- indicates an enchant
						end
						
						local reagentString = nil
						
						
						local reagentData = {}
		

						for j=1, API.GetNumReagents(i), 1 do
							local reagentName, _, numNeeded = API.GetReagentInfo(i,j)

							local reagentID = 0
							
							if reagentName then
								local reagentLink = API.GetReagentLink(i,j)

								reagentID = self:GetItemIDFromLink(reagentLink)
							else
								gotNil = true
								break
							end
							
							reagentData[j] = {}
							
							reagentData[j].id = reagentID
							reagentData[j].numNeeded = numNeeded
							
							if reagentString then
								reagentString = reagentString..":"..reagentID..":"..numNeeded
							else
								reagentString = reagentID..":"..numNeeded
							end
							
							self:ItemDataAddUsedInRecipe(reagentID,recipeID)				-- add a cross reference for where a particular item is used
						end
						
						recipe.reagentData = reagentData
						
						if gotNil then
							self.recacheRecipe[recipeID] = true
						else
							recipeString = tradeID.." "..itemString.." "..reagentString
							
							if #tools then
								recipeString = recipeString.." "..toolString
							end
							
							recipeDB[recipeID] = recipeString
						end
						
					end
				end
			else
				gotNil = true
			end
		until true
		
		if gotNil and recipeID then
			self.recacheRecipe[recipeID] = true
		end
	end
	
	
	-- the only way i know how to automatically make enchant groups
	if tradeID == 7411 then
		self:ScanEnchantingGroups(mainGroup)
	
	end
	
	self:RecipeGroupConstructDBString(mainGroup)

DebugSpam("Scan Complete")
	
--	CloseTradeSkill()
		
	self:InventoryScan(true)
	self:CalculateCraftableCounts()
	self:SortAndFilterRecipes()
DebugSpam("all sorted")
	self.scanInProgress = false
	
	collectgarbage("collect")
	
	return true
--	AceEvent:TriggerEvent("Skillet_Scan_Complete", profession)
end


function Skillet:EnchantingRecipeSlotAssign(recipeID, slot)
	local recipeString = self.db.account.recipeDB[recipeID]
		
	local tradeID, itemString, reagentString, toolString = string.split(" ",recipeString)
	
	if itemString == "0" then
		itemString = "0:"..slot
		
		self.db.account.recipeDB[recipeID] = tradeID.." 0:"..slot.." "..reagentString.." "..toolString
		
		self.data.recipeList[recipeID].slot = slot
	end
end



local invSlotLookup = {
	["HEADSLOT"] = "HeadSlot",
	["NECKSLOT"] = "NeckSlot",
	["SHOULDERSLOT"] = "ShoulderSlot",
	["CHESTSLOT"] = "ChestSlot",
	["WAISTSLOT"] = "WaistSlot",
	["LEGSSLOT"] = "LegsSlot",
	["FEETSLOT"] = "FeetSlot",
	["WRISTSLOT"] = "WristSlot",
	["HANDSSLOT"] = "HandsSlot",
	["FINGER0SLOT"] = "Finger0Slot",
	["TRINKET0SLOT"] = "Trinket0Slot",
	["BACKSLOT"] =	"BackSlot",
	["ENCHSLOT_WEAPON"] = "MainHandSlot",
	["ENCHSLOT_2HWEAPON"] = "MainHandSlot",
	["SHIELDSLOT"] = "SecondaryHandSlot",
}



function Skillet:ScanEnchantingGroups(mainGroup)
	local groupList = {}
	
	if mainGroup then	
		local craftSlots = { GetCraftSlots() }
		
		self:RecipeGroupClearEntries(mainGroup)
		
		for i=1,#craftSlots do
			local groupName
			local slotName = getglobal(craftSlots[i])
			
			local invSlot
			
			if groupList[slotName] then
				groupList[slotName] = groupList[slotName]+1
				groupName = slotName.." "..groupList[slotName]
			else
				groupList[slotName] = 1
				groupName = slotName
			end
					
			local currentGroup = self:RecipeGroupNew(self.currentPlayer, 7411, "Blizzard", groupName)			-- 7411 = enchanting
			
			SetCraftFilter(i+1)
			
			for s=1,GetNumCrafts() do
				local recipeLink = GetCraftRecipeLink(s)
				local recipeID = self:GetItemIDFromLink(recipeLink)
				
				if craftSlots[i] ~= "NONEQUIPSLOT" then
					invSlot = GetInventorySlotInfo(invSlotLookup[craftSlots[i]])
					self:EnchantingRecipeSlotAssign(recipeID, invSlot)
				end
				
DebugSpam("adding "..(recipeLink or "nil").." to "..groupName)				
				self:RecipeGroupAddRecipe(currentGroup, recipeID, self.data.skillIndexLookup[self.currentPlayer][recipeID])
			end
			
			self:RecipeGroupAddSubGroup(mainGroup, currentGroup, i)
		end
	end
	
	SetCraftFilter(1)
end

